@@ -1,28 +1,25 @@
 ## Config::Versioned
 ##
-## Written 2011 by Scott T. Hardin for the OpenXPKI project
-## Copyright (C) 2010, 2011 by The OpenXPKI Project
+## Written 2011-2012 by Scott T. Hardin for the OpenXPKI project
+## Copyright (C) 2010-2012 by The OpenXPKI Project
 ##
-## Based on the CPAN module App::Options
+## Was based on the CPAN module App::Options, but the import() stuff
+## bit me so we're turning into a Moose.
 ##
 ## vim: syntax=perl
 
 package Config::Versioned;
 
-use strict;
-use warnings;
+use Moose;
+use namespace::autoclean;
 
 =head1 NAME
 
 Config::Versioned - Simple, versioned access to configuration data
 
-=head1 VERSION
-
-Version 0.02
-
 =cut
 
-our $VERSION = '0.02';
+our $VERSION = '0.11';
 
 use Carp;
 use Config::Std;
@@ -31,18 +28,30 @@ use DateTime;
 use Git::PurePerl;
 use Path::Class;
 
-my $delimiter   = '.';
-my $delimiter_regex = qr/ \. /xms;
+has 'path' => ( is => 'ro', isa => 'ArrayRef', default => sub { [qw( . )] } );
+has 'filename' => ( is => 'ro', isa => 'Str' );
+has 'dbpath' =>
+  ( is => 'ro', default => 'cfgver.git', required => 1 );
+has 'author_name' => ( is => 'ro', isa => 'Str', default => "process: $@" );
+has 'author_mail' => (
+    is      => 'ro',
+    isa     => 'Str',
+    default => $ENV{GIT_AUTHOR_EMAIL} || $ENV{USER} . '@localhost'
+);
+has 'autocreate' => ( is => 'ro', isa => 'Bool', default => 0 );
+has 'commit_time' => ( is => 'ro', isa => 'DateTime' );
+has 'comment'     => ( is => 'rw', isa => 'Str' );
+has 'delimiter'   => ( is => 'ro', isa => 'Str', default => '.' );
+has 'delimiter_regex' =>
+  ( is => 'ro', isa => 'RegexpRef', default => sub { qr{ \. }xms } );
+has 'log_get_callback' => ( is => 'ro' );
+has '_git'             => ( is => 'rw'  );
+has 'debug'            => ( is => 'rw', isa => 'Int', default => 0 );
 
 # a reference to the singleton Config::Versioned object that parsed the command line
-my ($default_option_processor);
+#my ($default_option_processor);
 
-my (%path_is_secure);
-
-my $debug = 0;
-if ( defined $ENV{CONFIG_VERSIONED_DEBUG} ) {
-    $debug = $ENV{CONFIG_VERSIONED_DEBUG};
-}
+#my (%path_is_secure);
 
 =head1 SYNOPSIS
 
@@ -53,9 +62,6 @@ if ( defined $ENV{CONFIG_VERSIONED_DEBUG} ) {
     my $old1 = $cfg->get('subsystem1.group.param1', $version);
     my @keys = $cfg->list('subsys1.db');
 
-    my $cfg2 = Config::Versioned->new( { prefix => 'subsystem1.group' } );
-    my $p1 = $cfg2->get('param1');
-    my $p2 = $cfg2->get('param2');
 
 =head1 DESCRIPTION
 
@@ -133,20 +139,23 @@ commits.
 =item autocreate
 
 If no internal git repository exists, it will be created during code
-initialization. 
+initialization. Note that if an import filename is specified, this 
+automatically sets autocreate to true.
 
  autocreate => 1,
 
 The default is "0".
 
 Note: this option might become deprecated. I just wanted some extra
-"insurance" during the early stages of development.
+"insurance" during the early stages of development. 
 
 =item commit_time
 
 This sets the time to use for the commits in the internal git repository.
 It is used for debugging purposes only!
 
+Note: this must be a DateTime object instance.
+
 =item delimiter
 
 Specifies the delimiter used to separate the different levels in the
@@ -158,79 +167,76 @@ Specifies the delimiter used to separate the different levels in the
 string used to designate the location of a configuration parameter.
 [Default: qr/ \. /xms]
 
+=item log_get_callback
+
+Specifies a callback function to be called by get() after fetching
+the value for the given key. The subroutine should accept the
+parameters LOCATION, VERSION, VALUE. The VALUE may either be a single
+scalar value or an array reference containing a list of values.
+
+    sub cb_log_get {
+        my $self = shift;
+        my $loc = shift;
+        my $ver = shift;
+        my $val = shift;
+
+        warn "Access config parameter: $loc ($ver) => ",
+            ref($val) eq 'ARRAY'
+                ? join(', ', @{ $val })
+                : $val,
+            "\n";
+    }
+    my $cfg = Config::Versioned->new( { log_get_callback => 'cb_log_get' } );
+
+Note: if log_get_callback is a code ref, it will be called as a function.
+Otherwise, the log_get_callback will specify a method name that is to be
+called on the current object instance.
+
 =back
 
-=head2 new( { PARAMS } )
+=head2 BUILD( { PARAMS } )
 
-This is called during init() and creates an object instance. It may
-also be called elsewhere and given a I<prefix> to restrict C<get()>
-requests to a specific node.
+NOTE: This is used internally, so the typical user shouldn't bother with this.
+
+This is called after an object is created. When cloning, it is important that
+the new instance gets a reference to the same Git::PurePerl instance. This
+will prevent two instances from getting out of sync if modifications are made
+to the configuration data at runtime. To handle this, the parameter 'GITREF'
+must be passed when cloning.
+
+Note 2: this should be handled automatically in the _near_ future.
+
+    my $cv2 = $cv1->new( GITREF => $cv1->_git() );
 
 =cut
 
-sub new {
-    my ($this) = shift;
-    my $class  = ref($this) || $this;
-    my $self   = {};
-    my $params = shift;
+sub BUILD {
+    my $self = shift;
+    my $args = shift;
 
-    warn "# new() called with params: ", join( ', ', %{$params} ), "\n"
-      if $debug;
-    $Config::Versioned::init_args ||= {};
-    warn "# new() called with init_args: ",
-      join( ', ', %{$Config::Versioned::init_args} ), "\n"
-      if $debug;
-    bless $self, $class;
-
-    my $init_args = $Config::Versioned::init_args;
-
-    # process class args
-    foreach my $key (
-        qw( path filename dbpath directory autocreate author_name author_mail commit_time )
-      )
-    {
-        if ( exists $params->{$key} ) {
-            $init_args->{$key} = $params->{$key};
-        }
-    }
-    $self->{init_args} = $init_args;    # deprecated?!?
-
-    # process instance args
-
-    foreach my $key (qw( prefix commit_time author_name author_mail )) {
-        if ( exists $params->{$key} ) {
-            $self->{$key} = $params->{$key};
-        }
+    if ( defined $ENV{CONFIG_VERSIONED_DEBUG} ) {
+        $self->debug( $ENV{CONFIG_VERSIONED_DEBUG} );
     }
 
-    if ( not defined $Config::Versioned::git ) {
-        if ( not $self->_init_repo($init_args) ) {
-            return;
-        }
+    if ( not $self->_init_repo() ) {
+        return;
     }
-
-    my $cfghash = $self->parser($init_args);
+#    if ( not $self->_git() ) {
+#        if ( $args->{GITREF} ) {
+#            $self->_git( $args->{GITREF} );
+#        }
+#        else {
+#            if ( not $self->_init_repo() ) {
+#                return;
+#            }
+#        }
+#    }
+#
+#    $self->parser($args);
 
     return ($self);
 }
 
-=head2 isParam( LOCATION [, VERSION ] )
-
-Returns true if the given LOCATION contains a parameter value and false
-if it is a tree object containing subtrees or objects. In other words,
-if you get a true value on a LOCATION, you can fetch the value of the 
-parameter. Otherwise, you should use list() to see the names of the 
-locations subordinate to this LOCATION.
-
-NOTE: I don't like the wording of this, but it is more a
-minor kludge in a proof-of-concept implementation.
-
-=cut
-
-sub isParam {
-    die "isParam not implemented";
-}
-
 =head2 get( LOCATION [, VERSION ] )
 
 This is the accessor for fetching the value(s) of the given parameter. The
@@ -248,18 +254,23 @@ sub get {
     my $self     = shift;
     my $location = shift;
     my $version  = shift;
-
-    if ( $self->{prefix} ) {
-        $location = $self->{prefix} . $delimiter . $location;
-    }
-    my $obj = $self->_findobj( $location, $version );
+    my $cb       = $self->log_get_callback();
+    my ( $obj, $deobj ) = $self->_findobjx( $location, $version );
 
     if ( not defined $obj ) {
+        $self->$cb( $location, $version, '<undefined>' ) if $cb;
         return;
     }
 
     if ( $obj->kind eq 'blob' ) {
-        return $obj->content;
+        $self->$cb( $location, $version, $obj->content ) if $cb;
+        if ( $deobj->mode() == 120000 ) {
+            my $tmp = $obj->content;
+            return \$tmp;
+        }
+        else {
+            return $obj->content;
+        }
     }
     elsif ( $obj->kind eq 'tree' ) {
         my @entries = $obj->directory_entries;
@@ -267,11 +278,16 @@ sub get {
         foreach my $de (@entries) {
             push @ret, $de->filename;
         }
-        return
+        my @sorted =
           sort { ( $a =~ /^\d+$/ and $b =~ /^\d+$/ ) ? $a <=> $b : $a cmp $b }
           @ret;
+        $self->$cb( $location, $version, \@sorted ) if $cb;
+        return @sorted;
     }
     else {
+        $self->$cb( $location, $version,
+            "<error: non-blob object '" . $obj->kind . "' not supported>" )
+          if $cb;
         warn "# DEBUG: get() was asked to return a non-blob object [kind=",
           $obj->kind, "]\n";
         return;
@@ -312,9 +328,6 @@ sub kind {
     my $location = shift;
     my $version  = shift;
 
-    if ( $self->{prefix} ) {
-        $location = $self->{prefix} . $delimiter . $location;
-    }
     my $obj = $self->_findobj( $location, $version );
 
     if ( not defined $obj ) {
@@ -348,10 +361,6 @@ sub listattr {
     my $location = shift;
     my $version  = shift;
 
-    if ( $self->{prefix} ) {
-        $location = $self->{prefix} . $delimiter . $location;
-    }
-
     my $obj = $self->_findobj( $location, $version );
     if ( $obj and $obj->kind eq 'tree' ) {
         my @entries = $obj->directory_entries;
@@ -377,22 +386,18 @@ and returns a hashref to a named-parameter list.
 sub dumptree {
     my $self    = shift;
     my $version = shift;
-    my $cfg     = $Config::Versioned::git;
+    my $cfg     = $self->_git();
 
     # If no version hash was given, default to the HEAD of master
 
     if ( not $version ) {
-        if ( $cfg->all_sha1s->all ) {
-            my $master = $cfg->ref('refs/heads/master');
-            if ( not $master ) {
-                die "ERR: no master object found";
-            }
+        my $master = $self->_git()->ref('refs/heads/master');
+        if ( $master ) {
             $version = $master->sha1;
+        } else {
+            # if no sha1s are in repo, there's nothing to return
+            return;
         }
-        else {
-            return;    # if no sha1s are in repo, there's nothing to return
-        }
-
     }
 
     my $obj = $cfg->get_object($version);
@@ -418,7 +423,8 @@ sub dumptree {
         if ( $child->kind eq 'tree' ) {
             my $subret = $self->dumptree( $de->sha1 );
             foreach my $key ( keys %{$subret} ) {
-                $ret->{ $de->filename . $delimiter . $key } = $subret->{$key};
+                $ret->{ $de->filename . $self->delimiter() . $key } =
+                  $subret->{$key};
             }
         }
         elsif ( $child->kind eq 'blob' ) {
@@ -430,15 +436,6 @@ sub dumptree {
 
     }
     return $ret;
-
-    #            if ( $de->kind eq 'tree'
-    #            if ( $de->filename eq $key ) {
-    #                $found++;
-    #                $obj = $cfg->get_object( $de->sha1 );
-    #                last;
-    #            }
-    #        }
-
 }
 
 =head2 version
@@ -454,7 +451,7 @@ value if it is found.
 sub version {
     my $self    = shift;
     my $version = shift;
-    my $cfg     = $Config::Versioned::git;
+    my $cfg     = $self->_git();
 
     if ($version) {
         my $obj = $cfg->get_object($version);
@@ -473,112 +470,60 @@ sub version {
 
 =head1 INTERNALS
 
-=cut
+=head2 _init_repo
 
-# This translates the procedural Config::Versioned::import() into the class
-# method Config::Versioned->_import() (for subclassing)
+Initializes the internal git repository used for storing the config
+values. 
 
-sub import {
-    my ( $package, $args ) = @_;
-    $package->_import($args);
-}
+If the I<objects> directory in the C<dbpath> does not exist, an
+C<init()> on the C<Git::PurePerl> class is run. Otherwise, the 
+instance is initialized using the existing bare repository.
 
-sub _import_test {
-    my ( $class, $args ) = @_;
-    $default_option_processor = undef;
-    $class->_import($args);
-}
-
-sub _import {
-    my ( $class, $args ) = @_;
-
-    # We only do this once (the default Config::Versioned option processor
-    # is a singleton)
-
-    if ( not $Config::Versioned::init_args ) {
-
-        if ( not $args ) {
-            $args = {};    # default to an empty hash
-        }
-        elsif ( ref($args) ne 'HASH' ) {
-            croak "Config::Versioned::import(): args must be a HASHREF";
-        }
-        my $init_args = $args;
-
-        #        }
-
-        if ( not defined $init_args->{path} ) {
-            $init_args->{path} = [qw( . )];
-        }
-        elsif ( ref( $init_args->{path} ) ne 'ARRAY' ) {
-            croak "Config::Versioned 'path' must be a reference to an ARRAY";
-        }
-
-        #        if ( not $init_args->{filename} ) {
-        #            $init_args->{filename} = 'cfgver.conf';
-        #        }
-
-        if ( not $init_args->{dbpath} ) {
-            $init_args->{dbpath} = 'cfgver.git';
-        }
-
-        if ( exists $init_args->{delimiter} ) {
-            $delimiter = $init_args->{delimiter};
-        }
-        if ( exists $init_args->{delimiter_regex} ) {
-            $delimiter_regex = $init_args->{delimiter_regex};
-        }
-
-        $Config::Versioned::init_args = $init_args;
-
-        #        my $option_processor = $class->new($init_args);
-        #
-        #$default_option_processor =
-        #  $option_processor;    # save it in the singleton location
-
-    }
-}
-
-=head2 _init_repo INITARGS
+On error, it returns C<undef> and the reason is in C<$@>.
 
 =cut
 
 sub _init_repo {
-    my $self      = shift;
-    my $init_args = shift;
+    my $self = shift;
 
     my $git;
-    if ( not $init_args->{dbpath} ) {
-        die "ERROR: dbpath not set";
-    }
 
-    if ( not -d $init_args->{dbpath} ) {
-        if ( not dir( $init_args->{dbpath} )->mkpath ) {
-            $@ = 'Error creating directory ' . $init_args->{dbpath} . ': ' . $!;
-            return;
+    #    if ( not $init_args->{dbpath} ) {
+    #        die "ERROR: dbpath not set";
+    #    }
+
+    if ( not -d $self->dbpath() . '/objects' ) {
+        if ( $self->filename() || $self->autocreate() ) {
+            if ( not -d $self->dbpath() ) {
+                if ( not dir( $self->dbpath() )->mkpath ) {
+                    die 'Error creating directory ' . $self->dbpath() . ': ' . $!;
+                }
+            }
+            $git = Git::PurePerl->init( gitdir => $self->dbpath() );
+        } else {
+            die 'Error: dbpath (' . $self->dbpath() . ') does not exist';
         }
-        $git = Git::PurePerl->init( gitdir => $init_args->{dbpath} );
     }
     else {
-        $git = Git::PurePerl->new( gitdir => $init_args->{dbpath} );
+        $git = Git::PurePerl->new( gitdir => $self->dbpath() );
     }
-    $Config::Versioned::git = $git;
-
+    $self->_git($git);
+    $self->parser();
+    return $self;
 }
 
-=head2 _import_cfg INITARGS
+=head2 _get_anon_scalar
 
-This wrapper calls parser() and is provided for
-backward compatibility.
+Creates an anonymous scalar for representing symlinks in the tree structure.
 
 =cut
 
-sub _import_cfg {
-    my $self    = shift;
-    my $cfghash = $self->parser(@_);
+sub _get_anon_scalar {
+    my $temp = shift;
+    return \$temp;
 }
 
-=head2 parser INITARGS
+=head2 parser ARGS
 
 Imports the configuration read and writes it to the internal database. If no
 filename is passed as an argument, then it will quietly skip the commit.
@@ -589,8 +534,8 @@ is a simple example:
 
     sub parser {
         my $self = shift;
-        my $params = shift;
-        $params->{comment} = 'import from my perl hash';
+        my $args = shift;
+        $args->{comment} = 'import from my perl hash';
         
         my $cfg = {
             group1 => {
@@ -605,10 +550,16 @@ is a simple example:
                     param4 => 'val4',
                 },
             },
+            # This creates a symlink from 'group3.subgroup3' to 'connector1/group4'.
+            # Note the use of the scalar reference using the backslash.
+            group3 => {
+                subgroup3 => \'connector1/group4',
+            },
+
         };
         
-        # pass original params, appended with a comment string for the commit
-        $self->commit( $cfg, $params );
+        # pass original args, appended with a comment string for the commit
+        $self->commit( $cfg, $args );
     }
 
 In the comment, you should include details on where the config came from
@@ -617,32 +568,31 @@ In the comment, you should include details on where the config came from
 =cut
 
 sub parser {
-    my $self   = shift;
-    my $params = shift;
+    my $self = shift;
+    my $args = shift;
 
-    # populate our local params with init_args, if needed,
-    # without overwriting init_args
-    foreach my $key ( keys %{$Config::Versioned::init_args} ) {
-        if ( not exists $params->{$key} ) {
-            $params->{$key} = $Config::Versioned::init_args->{$key};
+    foreach
+      my $key (qw( comment filename path author_name author_mail commit_time ))
+    {
+        if ( not exists $args->{$key} ) {
+            $args->{$key} = $self->$key();
         }
     }
 
     # If no filename was specified, then there is no import of
     # configuration files needed. Quietly exit method.
 
-    if ( not $params->{filename} ) {
+    if ( not $args->{filename} ) {
         return $self;
     }
 
     # Read the configuration from the import files
 
     my %cfg = ();
-    $self->_read_config_path( $params->{filename}, \%cfg,
-        @{ $params->{path} } );
+    $self->_read_config_path( $args->{filename}, \%cfg, @{ $args->{path} } );
 
-    my $comment = "Import config from "
-      . $self->_which( $params->{filename}, @{ $params->{path} } );
+    $args->{comment} ||= "Import config from "
+      . $self->_which( $args->{filename}, @{ $args->{path} } );
 
     # convert the foreign data structure to a simple hash tree,
     # where the value is either a scalar or a hash reference.
@@ -652,7 +602,7 @@ sub parser {
 
         # build up the underlying branch for these leaves
 
-        my @sectpath = split( $delimiter_regex, $sect );
+        my @sectpath = split( $self->delimiter_regex(), $sect );
         my $sectref = $tmphash;
         foreach my $nodename (@sectpath) {
             $sectref->{$nodename} ||= {};
@@ -662,55 +612,61 @@ sub parser {
         # now add the leaves
 
         foreach my $leaf ( keys %{ $cfg{$sect} } ) {
-            $sectref->{$leaf} = $cfg{$sect}{$leaf};
+
+            # If the leaf start or ends with an '@', treat it as
+            # a symbolic link.
+            if ( $leaf =~
+                m{ (?: \A @ (.*?) @ \z | \A @ (.*) | (.*?) @ \z ) }xms )
+            {
+                my $match = $1 || $2 || $3;
+
+                # make it a ref to an anonymous scalar so we know it's a symlink
+                #my $t = _get_anon_scalar($1);
+                $sectref->{$match} = \( $cfg{$sect}{$leaf} );
+            }
+            else {
+                $sectref->{$leaf} = $cfg{$sect}{$leaf};
+            }
         }
 
     }
 
-    $params->{comment} = $comment;
-    $self->commit( $tmphash, $params );
+    $self->commit( $tmphash, $args );
 }
 
-=head2 commit CFGHASH, INITARGS
+=head2 commit CFGHASH[, ARGS]
 
 Import the configuration tree in the CFGHASH anonymous hash and commit
 the modifications to the internal git bare repository.
 
+ARGS is a ref to a named-parameter list (e.g. HASH) that may contain the
+following keys to override the instance defaults:
+
+    author_name, author_mail, comment, commit_time
+
 =cut
 
 sub commit {
     my $self = shift;
     my $hash = shift;
+    my $args = shift;
 
     if ( ref($hash) ne 'HASH' ) {
         confess "ERR: commit() - arg not hash ref [$hash]";
     }
 
-    my $params = shift;
-
-    # populate our local params with init_args, if needed,
-    # without overwriting init_args
-    foreach my $key ( keys %{$Config::Versioned::init_args} ) {
-        if ( not exists $params->{$key} ) {
-            $params->{$key} = $Config::Versioned::init_args->{$key};
-        }
-    }
-
     my $parent = undef;
     my $master = undef;
 
-    if ( $Config::Versioned::git->all_sha1s->all ) {
-        $master = $Config::Versioned::git->ref('refs/heads/master');
-        if ( not $master ) {
-            die "ERR: no master object found";
-        }
+    $master = $self->_git()->ref('refs/heads/master');
+    if ( $master ) {
         $parent = $master->sha1;
     }
 
-    #    warn "# author_name: ", $init_args->{author_name}, "\n";
+    #    warn "# author_name: ", $self->author_name(), "\n";
     my $tree = $self->_hash2tree($hash);
 
-    if ($debug) {
+    if ( $self->debug() ) {
         print join( "\n# ", '', $self->_debugtree($tree) ), "\n";
     }
 
@@ -722,7 +678,7 @@ sub commit {
     #
 
     if ( $parent and $master->tree->sha1 eq $tree->sha1 ) {
-        if ($debug) {
+        if ( $self->debug() ) {
             carp("Nothing to commit (index matches HEAD)");
         }
         return $self;
@@ -733,11 +689,11 @@ sub commit {
     #
 
     my $actor = Git::PurePerl::Actor->new(
-        name  => $params->{author_name} || "process: $0",
-        email => $params->{author_mail} || $ENV{USER} . '@localhost',
+        name  => $args->{author_name} || $self->author_name,
+        email => $args->{author_mail} || $self->author_mail,
     );
 
-    my $time = $params->{commit_time} || DateTime->now;
+    my $time = $args->{commit_time} || $self->commit_time || DateTime->now;
 
     my @commit_attrs = (
         tree           => $tree->sha1,
@@ -745,14 +701,14 @@ sub commit {
         authored_time  => $time,
         committer      => $actor,
         committed_time => $time,
-        comment        => $params->{comment} || 'import config',
+        comment        => $args->{comment} || $self->comment(),
     );
     if ($parent) {
         push @commit_attrs, parent => $parent;
     }
 
     my $commit = Git::PurePerl::NewObject::Commit->new(@commit_attrs);
-    $Config::Versioned::git->put_object($commit);
+    $self->_git()->put_object($commit);
 
 }
 
@@ -763,57 +719,90 @@ sub _hash2tree {
     if ( ref($hash) ne 'HASH' ) {
         confess "ERR: _hash2tree() - arg not hash ref [$hash]";
     }
-    if ($debug) {
+    if ( $self->debug() ) {
         warn "Entered _hash2tree( $hash ): ", join( ', ', %{$hash} ), "\n";
     }
 
     my @dir_entries = ();
 
     foreach my $key ( keys %{$hash} ) {
-        if ($debug) {
+        if ( $self->debug() ) {
             warn "# _hash2tree() processing $key -> ", $hash->{$key}, "\n";
         }
         if ( ref( $hash->{$key} ) eq 'HASH' ) {
-            if ($debug) {
+            if ( $self->debug() ) {
                 warn "# _hash2tree() adding subtree for $key\n";
             }
             my $subtree = $self->_hash2tree( $hash->{$key} );
+             
+            next unless($subtree);
+
+            my $local_key = $key;
+            if ( $] > 5.007 && utf8::is_utf8($local_key) ) {
+                utf8::downgrade($local_key);
+            }
+
             my $de      = Git::PurePerl::NewDirectoryEntry->new(
                 mode     => '40000',
-                filename => $key,
-                sha1     => $subtree->sha1,
+                filename => $local_key,
+                sha1     => $subtree->sha1(),
             );
             push @dir_entries, $de;
         }
-        else {
+        elsif ( ref( $hash->{$key} ) eq 'SCALAR' ) {
+
+            # Support for symbolic links
+            if ( $self->debug() ) {
+                warn "# _hash2tree() adding symlink for $key\n";
+            }
+            my $obj =
+              Git::PurePerl::NewObject::Blob->new(
+                content => ${ $hash->{$key} } );
+            $self->_git()->put_object($obj);
+            my $local_key = $key;
+            if ( $] > 5.007 && utf8::is_utf8($local_key) ) {
+                utf8::downgrade($local_key);
+            }
+            my $de = Git::PurePerl::NewDirectoryEntry->new(
+                mode     => '120000',     # symlink
+                filename => $local_key,
+                sha1     => $obj->sha1(),
+            );
+            push @dir_entries, $de;
+        }
+        elsif ( defined $hash->{$key} ) {
             my $obj =
               Git::PurePerl::NewObject::Blob->new( content => $hash->{$key} );
-            $Config::Versioned::git->put_object($obj);
+
+            my $local_key = $key;
+            if ( $] > 5.007 && utf8::is_utf8($local_key) ) {
+                utf8::downgrade($local_key);
+            }
+
+            warn "# created blob for '$key' with sha " . $obj->sha1() if $self->debug();
+            warn "#      '$key' utf8 flag: ", utf8::is_utf8($key) if $self->debug();
+            $self->_git()->put_object($obj);
             my $de = Git::PurePerl::NewDirectoryEntry->new(
-                mode     => '100644',
-                filename => $key,
-                sha1     => $obj->sha1,
+                mode     => '100644',     # plain file
+                filename => $local_key,
+                sha1     => $obj->sha1(),
             );
             push @dir_entries, $de;
+        } else {
+            warn "#  _hash2tree() value is undef for key $key\n";
         }
     }
+
+    if (!scalar @dir_entries) {
+        warn "# _hash2tree() nothing to push\n";
+        return undef;
+    }
+
     my $tree =
       Git::PurePerl::NewObject::Tree->new( directory_entries =>
           [ sort { $a->filename cmp $b->filename } @dir_entries ] );
 
-#    my $sha1 = Digest::SHA->new();
-#    $sha1->add( $tree->raw() );
-#    my $test_sha1 = $sha1->hexdigest();
-#    if ( $test_sha1 ne $tree->sha1 ) {
-#        warn "WARNING: my sha1 $test_sha1; Git::PurePerl sha1 ", $tree->sha1, "\n";
-#    }
-#        if ( $test_sha1 eq 'c2b1cf11f2abf788bfef75bbdf0263c84c3eb058' or $test_sha1 eq 'deff7d65d491fe1291323a82080bb94d72577ac7' ) {
-#            warn "HEY!!!\n";
-#            warn hdump( $tree->raw ), "\n";
-#        }
-#    }
-
-    if ($debug) {
+    if ( $self->debug() ) {
         my $content = $tree->content;
         $content =~ s/(.)/sprintf("%x",ord($1))/eg;
         warn "# Added tree with dir entries: ",
@@ -825,7 +814,7 @@ sub _hash2tree {
 
     }
 
-    $Config::Versioned::git->put_object($tree);
+    $self->_git()->put_object($tree);
 
     return $tree;
 }
@@ -841,8 +830,8 @@ A reference to the node at the LOCATION is returned.
 sub _mknode {
     my $self     = shift;
     my $location = shift;
-    my $ref      = $Config::Versioned::git;
-    foreach my $key ( split( $delimiter_regex, $location ) ) {
+    my $ref      = $self->_git();
+    foreach my $key ( split( $self->delimiter_regex(), $location ) ) {
         if ( not exists $ref->{$key} ) {
             $ref->{$key} = {};
         }
@@ -856,33 +845,34 @@ sub _mknode {
     return $ref;
 }
 
-=head2 _findobj LOCATION [, VERSION ]
+=head2 _findobjx LOCATION [, VERSION ]
 
-Returns the Git::PurePerl object found in the file path at LOCATION.
+Returns the Git::PurePerl and Git::PurePerl::DirectoryEntry objects found in
+the file path at LOCATION.
 
-    my $ref1 = $cfg->_findnode("smartcard.ldap.uri");
-    my $ref2 = $cfg->_findnode("certs.signature.duration", $wfcfgver);
+    my ($ref1, $de1) = $cfg->_findnode("smartcard.ldap.uri");
+    my $ref2, $de2) = $cfg->_findnode("certs.signature.duration", $wfcfgver);
+
+In most cases, the C<_findobj> version is sufficient. This extended version
+is used to look at the attribtes of the directory entry for things like whether
+the blob is a symlink.
 
 =cut
 
-sub _findobj {
+sub _findobjx {
     my $self     = shift;
     my $location = shift;
     my $ver      = shift;
-    my $cfg      = $Config::Versioned::git;
+    my $cfg      = $self->_git();
+    my ( $obj, $deobj );
 
     # If no version hash was given, default to the HEAD of master
 
     if ( not $ver ) {
-        if ( $Config::Versioned::git->all_sha1s->all ) {
-            my $master = $Config::Versioned::git->ref('refs/heads/master');
-            if ( not $master ) {
-                die "ERR: no master object found";
-            }
+        my $master = $self->_git()->ref('refs/heads/master');
+        if ( $master ) {
             $ver = $master->sha1;
-        }
-        else {
-
+        } else {
             # if no sha1s are in repo, there's nothing to return
             return;
         }
@@ -892,7 +882,7 @@ sub _findobj {
     # TODO: is this the way we want to handle the error of not finding
     # the given object?
 
-    my $obj = $cfg->get_object($ver);
+    $obj = $cfg->get_object($ver);
     if ( not $obj ) {
         $@ = "No object found for SHA1 $ver";
         return;
@@ -901,13 +891,18 @@ sub _findobj {
     if ( $obj->kind eq 'commit' ) {
         $obj = $obj->tree;
     }
-    my @keys = split $delimiter_regex, $location;
+    my @keys = split $self->delimiter_regex(), $location;
 
     # iterate thru the levels in the location
 
     while (@keys) {
         my $key = shift @keys;
 
+        # if the object is a blob, we already reached the leaf
+        if ($obj->kind eq 'blob') {
+            return undef;
+        }
+
         # $obj should contain the parent tree object.
 
         my @directory_entries = $obj->directory_entries;
@@ -918,7 +913,8 @@ sub _findobj {
         foreach my $de (@directory_entries) {
             if ( $de->filename eq $key ) {
                 $found++;
-                $obj = $cfg->get_object( $de->sha1 );
+                $obj   = $cfg->get_object( $de->sha1 );
+                $deobj = $de;
                 last;
             }
         }
@@ -927,15 +923,34 @@ sub _findobj {
             return;
         }
     }
-    return $obj;
+    return $obj, $deobj;
 
 }
 
+=head2 _findobj LOCATION [, VERSION ]
+
+Returns the Git::PurePerl object found in the file path at LOCATION.
+
+    my $ref1 = $cfg->_findnode("smartcard.ldap.uri");
+    my $ref2 = $cfg->_findnode("certs.signature.duration", $wfcfgver);
+
+=cut
+
+sub _findobj {
+    my $self = shift;
+    my ( $obj, $deobj ) = $self->_findobjx(@_);
+    if ( defined $obj ) {
+        return $obj;
+    }
+    else {
+        return;
+    }
+}
+
 =head2 _get_sect_key LOCATION
 
 Returns the section and key needed by Config::Std to access the
-configuration values. The given LOCATION is prepended with the 
-current prefix, if set, and is split on the last delimiter. 
+configuration values. The given LOCATION is split on the last delimiter. 
 The resulting section and key are returned as a list.
 
 =cut
@@ -943,30 +958,17 @@ The resulting section and key are returned as a list.
 sub _get_sect_key {
     my $self = shift;
     my $key  = shift;
-    if ( $self->{prefix} ) {
-        $key = $self->{prefix} . $delimiter . $key;
-    }
 
     # Config::Std uses section/key, so we need to split up the
     # given key
 
-    my @tokens = split( $delimiter_regex, $key );
+    my @tokens = split( $self->delimiter_regex(), $key );
     $key = pop @tokens;
-    my $sect = join( $delimiter, @tokens );
+    my $sect = join( $self->delimiter(), @tokens );
 
     return $sect, $key;
 }
 
-=head2 _read_options()
-
-Called during init(), this method reads the actual configuration values
-and populates the internal data structure.
-
-=cut
-
-sub _read_options {
-}
-
 =head2 _which( NAME, DIR ... )
 
 Searches the directory list DIR, returning the full path in which the file NAME was
@@ -1023,25 +1025,16 @@ sub _debugtree {
     my $self   = shift;
     my $start  = shift;
     my $indent = shift || 0;
-    my $cfg    = $Config::Versioned::git;
+    my $cfg    = $self->_git();
     my @out    = ();
 
     my $tabsize = 2;
     my $obj;
 
-    if ($debug) {
-
-        #        warn "# Entered _debugtree( $start, ", $indent || 0, ")\n";
-    }
-
     # Soooo, let's see what we've been fed...
-
     if ( not $start ) {    # default to the HEAD of master
-        if ( $cfg->all_sha1s->all ) {
-            my $master = $cfg->ref('refs/heads/master');
-            if ( not $master ) {
-                die "ERR: no master object found";
-            }
+        my $master = $cfg->ref('refs/heads/master');
+        if ( $master ) {
             $obj = $cfg->get_object( $master->sha1 );
         }
         else {
@@ -1096,7 +1089,8 @@ sub _debugtree {
 
     if ( $obj->kind eq 'commit' ) {
         foreach my $attr (
-            qw( tree_sha1 parent_sha1s author authored_time committer commited_time comment encoding )
+            qw( tree_sha1 parent_sha1s author authored_time committer
+            commited_time comment encoding )
           )
         {
             if ( $obj->can($attr) ) {
@@ -1114,13 +1108,6 @@ sub _debugtree {
             ( ' ' x ( $tabsize * $indent ) ) . $_
         } hdump( $obj->kind . ' ' . $obj->size . "\0" . $obj->content );
 
-   #         my $sha1 = Digest::SHA1->new;
-   #         $sha1->add( $obj->kind . ' ' . $obj->size . "\0" . $obj->content );
-   #
-   #        push @out,
-   #              ( ' ' x ( $tabsize * $indent ) )
-   #              . 'my sha1: '. $sha1->hexdigest;
-
         my $sha1a = Digest::SHA->new;
         $sha1a->add( $obj->kind . ' ' . $obj->size . "\0" . $obj->content );
 
@@ -1192,11 +1179,15 @@ sub hdump {
 
 =head1 ACKNOWLEDGEMENTS
 
-Based on the CPAN module App::Options.
+Was based on the CPAN module App::Options, but since been converted to Moose.
 
 =head1 AUTHOR
 
-Scott T. Hardin, C<< <mrscotty at hnsc.de> >>
+Scott T. Hardin, C<< <mrscotty at cpan.org> >>
+
+Martin Bartosch
+
+Oliver Welter
 
 =head1 BUGS
 
@@ -1245,5 +1236,7 @@ and/or modify it under the same terms as Perl itself.
 
 =cut
 
+__PACKAGE__->meta->make_immutable;
+
 1;    # End of Config::Versioned
 
